/*
 * Copyright 2017 Alexander Sychev
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     1. Redistributions of source code must retain the above copyright notice,
 *        this list of conditions and the following disclaimer.
 *     2. Redistributions in binary form must reproduce the above copyright
 *        notice, this list of conditions and the following disclaimer in the
 *        documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const glob = require("glob");
const path_1 = require("path");
const _ = require("lodash");
/** Cached "require" */
const __require = require;
/**
 * Common concrete implementation of module to autoload
 */
class Module {
    constructor(id, bootstrapInternal, opt_lazy) {
        this.id = id;
        this.lazy = opt_lazy || false;
        this.bootstrapInternal_ = bootstrapInternal;
    }
    /** Outer bootstrap wrapper method */
    bootstrap(context) {
        return this.bootstrapInternal_(context);
    }
}
exports.Module = Module;
/** Loader */
class Loader {
    /** Init loader by directories */
    constructor(dirs) {
        _.forEach(dirs, this.lintDir_);
        this.dirs_ = dirs ? _.uniq(_.compact(dirs)) : [];
    }
    /** Add directory to search modules */
    addDir(dir) {
        this.lintDir_(dir);
        this.dirs_.push(dir);
    }
    /** Find and load all modules into Application */
    loadModules() {
        return this.modulePathsToModules_(this.filePathsToModulesPaths_(this.globsToFilePaths_(this.dirsToGlobs_(this.dirs_))));
    }
    /** Map dirs array to Glob patterns array */
    dirsToGlobs_(dirs) {
        return _.map(dirs, (dir) => path_1.join(dir, '**', '*.auto.@(js|json|node)'));
    }
    /** Converts flat Glon patterns to modules paths*/
    globsToFilePaths_(globs) {
        return _.flatten(_.map(globs, (pattern) => glob.sync(pattern)));
    }
    /** Maps raw modules files paths into correct paths for "require" */
    filePathsToModulesPaths_(filePaths) {
        return _.map(filePaths, (filePath) => ((path_1.extname(filePath) === '.js') ?
            filePath.substring(0, filePath.length - 3) :
            filePath.substring(0, filePath.length - 5)));
    }
    /** Maps modules paths to modules */
    modulePathsToModules_(modulePaths) {
        return _.map(modulePaths, (modulePath) => __require(modulePath));
    }
    /** Lind directory path (check that path is absolute) */
    lintDir_(dir) {
        if (!path_1.isAbsolute(dir)) {
            throw new Error(`"${dir}" <== Root autoload directory path must be absolute!`);
        }
    }
}
exports.Loader = Loader;
;
/** Concrete context implementation */
class Context {
    /** Context requies instance of loader */
    constructor(loader) {
        this.notInitalized_ = new Map();
        this.cache_ = new Map();
        this.otherContexts_ = new Map();
        loader.loadModules().forEach(this.baseModuleInit_, this);
        this.bootstrapIfNotLazy_();
    }
    /** Put other context to this */
    putContext(id, context) {
        this.otherContexts_.set(id, context);
    }
    /** Get other context, puted in this context */
    getContext(id) {
        if (!this.otherContexts_.has(id)) {
            throw new Error(`Context has no parent context with id="${id}"`);
        }
        return this.otherContexts_.get(id);
    }
    /** Get exported module object by module id */
    getModule(id) {
        let result = null;
        if (this.cache_.has(id)) {
            result = this.cache_.get(id);
        }
        else if (this.notInitalized_.has(id)) {
            result = this.notInitalized_.get(id).bootstrap(this);
            this.cache_.set(id, result);
            this.notInitalized_.delete(id);
        }
        else {
            throw new Error(`Module with id="${id}" not registered by Autoloader`);
        }
        return result;
    }
    /** Base init of all modules */
    baseModuleInit_(module) {
        if (_.isNil(module.id)) {
            throw new Error('Autoloadable module must export "id" field at least!');
        }
        if (_.isNil(module.bootstrap)) {
            this.cache_.set(module.id, module);
        }
        else {
            this.notInitalized_.set(module.id, module);
        }
    }
    /** Automatic bootstrap module if it is not lazy */
    bootstrapIfNotLazy_() {
        for (let id of this.notInitalized_.keys()) {
            let isLazy = Boolean(this.notInitalized_.get(id).lazy);
            if (!isLazy)
                this.getModule(id);
        }
    }
}
exports.Context = Context;
